
#################################################
# Subroutines added by ST
#################################################

sub minidebug {
  my $dnowtime = strftime("%a, %d %b %Y %H:%M:%S %Z", localtime(time));
  print LOG "$dnowtime:$nprocess: ",@_,"\n" if ($MINIDEBUG && !$DEBUG);
}

sub close_log {
  ($sec,$min,$hour,$mday,$mon,$year) = localtime(time);

  &debug("--- all finished. Total of ",tv_interval ($start_time, [gettimeofday])," secs");
  &minidebug("------ Process $nprocess finished. Total of ",tv_interval ($start_time, [gettimeofday])," secs");
  close(LOG);
}

sub reject_email {
  my ($exit_string,$exit_code)=@_;
  $exit_code=111 if (!$exit_code);

  # st: tell qmail-smtpd why the message is rejected,
  # so it can be written to the qmail-smtpd log
  #warn "$nppid QS-$VERSION: $exit_string\n";
  warn "$V_HEADER-$VERSION: $exit_string\n";

  &debug("r_e: $V_HEADER-$VERSION: $exit_string");
  #&minidebug("r_e: QS-$VERSION: $exit_string");
  &minidebug("r_e: $V_HEADER-$VERSION: $exit_string");

  &cleanup;

  &close_log;
  exit $exit_code;
}

##############################################
# st: SCANNERS PER DOMAIN routines
##############################################

sub start_scanners {
  my($e_sender,$f_recips,$msg)=@_;
  $sa_rcpt='0';

  # Now, start the scanners!
  &init_scanners if ($scanner_array[0] ne "none");

  # st: if the message is marked to delete skip the mailing routines
  if (!$del_message) {
    if ($quarantine_event && ($scanner_array[0] ne "none")) {
      &debug("unsetting TCPREMOTEIP env var");
      delete $ENV{'TCPREMOTEIP'};
      #Reset locale back to original
      $ENV{'LC_ALL'}=$orig_locale;
      if ($sa_forward ne "" && $quarantine_event =~/spam/i && $description !~/potential virus/i) {
         if ($sa_fwd_verbose) {
            $sa_hdr_report='1' if ($sa_alt && $sa_debug && $sa_report);
            &qmail_parent_check;
            &qmail_requeue($e_sender,"T$sa_forward\0\0",$msg);
         } else {
            open (SF,"$qmailinject -f$returnpath $sa_forward < $msg|")||&error_condition("cannot run $qmailinject -f$returnpath $sa_forward < $msg - $!");
            close SF ;
         }
         # st: forward the messages just once..
         $sa_rcpt='0';
         $sa_forward='';
      }
      &email_quarantine_report;
    } else {
      &qmail_parent_check;
      &qmail_requeue($e_sender,$f_recips,$msg); 
    }
  }
}

sub scanners_p_d {
  my (%domain_scanners,@scanners_rcpt_array,%seen,$scanners_array);

  &debug("s_p_d: reading from $scanners_per_domain.db");
  tie (%domain_scanners,'DB_File',"$scanners_per_domain.db",O_RDONLY, 0600, $DB_HASH) || &error_condition("cannot open $scanners_per_domain.db - $!");

  # Check if we have a match within the database
  # Check order:
  # 1) return-path
  # 2) domain-return-path
  # 3) for each recipient: recipient, domain-recipient
  if ((exists $domain_scanners{$returnpath}) && (defined($ENV{'RELAYCLIENT'}))) {
     @scanner_array=split(/,/,$domain_scanners{"$returnpath"});
     &debug("s_p_d: return-path '$returnpath', scanners '$domain_scanners{$returnpath}'");
     &minidebug("s_p_d: return-path '$returnpath', scanners '$domain_scanners{$returnpath}'");
  }
  elsif ((exists $domain_scanners{$domain_returnpath}) && (defined($ENV{'RELAYCLIENT'}))) {
     @scanner_array=split(/,/,$domain_scanners{$domain_returnpath});
     &debug("s_p_d: domain-return-path match '$domain_returnpath', scanners '$domain_scanners{$domain_returnpath}'");
     &minidebug("s_p_d: domain-return-path match '$domain_returnpath', scanners '$domain_scanners{$domain_returnpath}'");
  }
  elsif ($one_recip && (exists $domain_scanners{$one_recip})) {
     @scanner_array=split(/,/,$domain_scanners{$one_recip});
     &debug("s_p_d: rcpt '$one_recip', scanners '$domain_scanners{$one_recip}'");
     &minidebug("s_p_d: rcpt '$one_recip', scanners '$domain_scanners{$one_recip}'");
  }
  elsif ($one_recip && (exists $domain_scanners{$domain_one_recip})) {
     @scanner_array=split(/,/,$domain_scanners{$domain_one_recip});
     &debug("s_p_d: domain_rcpt '$domain_one_recip', scanners '$domain_scanners{$domain_one_recip}'");
     &minidebug("s_p_d: domain_rcpt '$domain_one_recip', scanners '$domain_scanners{$domain_one_recip}'");
  }
  elsif (!$one_recip) {
     &debug("s_p_d: we have multiple recipient, checking each of them");
     &minidebug("s_p_d: we have multiple recipient, checking each of them");
     my @mrecips=split(',',$recips);
     my $mrcpt='';
     my $domain_mrcpt='';
     my %m_rcpt;
     foreach $mrcpt(@mrecips) {
       $mrcpt=tolower($mrcpt);
       $domain_mrcpt=$mrcpt;
       $domain_mrcpt=~ s/^(.*)\@(.*)$/$2/;
       if (exists $domain_scanners{$mrcpt}) {
          @scanner_array=split(/,/,$domain_scanners{$mrcpt});
       }
       elsif (exists $domain_scanners{$domain_mrcpt}) {
          @scanner_array=split(/,/,$domain_scanners{$domain_mrcpt});
       } else {
         @scanner_array=@scanners_default;
       }
       @scanner_array=&check_scanners(@scanner_array);
       $m_rcpt{$mrcpt}=join(',',@scanner_array);
     }
     untie %domain_scanners;
     while( ($one_recip,$scanners_array)=each %m_rcpt) {
       &debug("s_p_d: recipient '$one_recip', scanners '$scanners_array'");
       &minidebug("s_p_d: recipient '$one_recip', scanners '$scanners_array'");
       @scanner_array=split(',',$scanners_array);
       &start_scanners($env_returnpath,"T$one_recip\0\0","$scandir/$wmaildir/new/$file_id");
     }
     return;
  } else {
     @scanner_array=@scanners_default;
     &debug("s_p_d: no match, falling to the scanners_default");
     &minidebug("s_p_d: no match, falling to the scanners_default");
  }
  # if no multiples recipients
  untie %domain_scanners;
  @scanner_array=&check_scanners(@scanner_array);
  &start_scanners($env_returnpath,$env_recips,"$scandir/$wmaildir/new/$file_id");
}

sub generate_spd {
  my ($line,$count,%domain_scanners,$match_rcpt,$scanners_rcpt,@scanners_rcpt_array,%seen);

  print "\n  Generating $scanners_per_domain.db\n\n";

  unlink ("$scanners_per_domain.db.tmp");
  tie (%domain_scanners,'DB_File',"$scanners_per_domain.db.tmp",O_CREAT|O_RDWR,0640,$DB_HASH) || &error_condition("cannot open for write $scanners_per_domain.db.tmp - $!");
  #tie (%domain_scanners,'DB_File',"$scanners_per_domain.db.tmp",O_CREAT|O_RDWR,0640,$DB_BTREE) || &error_condition("cannot open for write $scanners_per_domain.db.tmp - $!");
  
  open(SPD, "<$scanners_per_domain.txt") ||&error_condition("cannot read $scanners_per_domain.txt - $!");;
  
  while (<SPD>) {
    $line++;
    next if (/^\#|^\s.*$/);    # Ignore lines starting with # or spaces
    next if (!(/:/));          # Ignore lines doesn't contain a ':'
    if (/\;|\!/) {
       print "d_w: line $line contains an invalid char, SKIP\n";
       next;
    }
    chomp;
    s/\s|\t//g;
    ($match_rcpt,$scanners_rcpt)=split(/:/,$_);
    $match_rcpt=tolower("$match_rcpt");
    $scanners_rcpt=tolower("$scanners_rcpt");
  
    if (exists $domain_scanners{$match_rcpt}) {
       print "  d_w: duplicated value '$match_rcpt' at line $line, SKIP \n";
       next;
    }
  
  
    # Let check if the scanner are really installed,
    # change 'sa' and 'ps' for the correct name, and
    # add _scanner to the AVs scanners
  
    @scanners_rcpt_array=split(/,/,$scanners_rcpt);
    foreach (@scanners_rcpt_array) {
       s/^sa$/spamassassin/;
       s/^ps$/perlscan/;
       s/^perlscanner$/perlscan/;
       s/^(.*)$/$1_scanner/ if((!/spamassassin/) && (!/_scanner/) && (!/^none$/));
    }
  
    # Check if the scanners are installed
    @scanners_rcpt_array=&check_scanners(@scanners_rcpt_array);
  
    $scanners_rcpt = join(',',@scanners_rcpt_array);
  
    # Check if at least we have one valid scanner
  
    if (@scanners_rcpt_array==0) {
       print "  d_w: There are no valid scanner for address '$match_rcpt' at line $line, SKIP\n";
       next;
    }
    $count++;
    # print " address: $match_rcpt / scanner: $scanners_rcpt\n";
    $domain_scanners{$match_rcpt}=$scanners_rcpt;
  }
  close(SPD);
  untie %domain_scanners;
  rename( "$scanners_per_domain.db.tmp", "$scanners_per_domain.db" );
  print "\n  Read $line lines, got $count entries\n\n";
}


sub read_spd {
  my ($count,%domain_scanners,$match_rcpt,$scanners_rcpt);
  $count=0;
  
  print "\n  Reading from $scanners_per_domain.db\n\n";
  #tie (%domain_scanners,'DB_File',"$scanners_per_domain.db",O_RDONLY, 0600, $DB_BTREE) || &error_condition("cannot open for write $scanners_per_domain.db - $!");;
  tie (%domain_scanners,'DB_File',"$scanners_per_domain.db",O_RDONLY, 0600, $DB_HASH) || &error_condition("cannot open for write $scanners_per_domain.db - $!");;
  
  foreach(sort keys %domain_scanners) {
     print " $_ : $domain_scanners{$_}\n";
     $count++;
  }
  
  untie %domain_scanners;
  print "\n d_w: total of $count entries found\n\n";
}


sub check_scanners {
  # Check against the installed scanners
  my @scanners_to_check=@_;
  return @scanners_to_check if ($scanners_to_check[0] eq "none");
  my %seen=();
  foreach (@scanners_installed) {
     $seen{$_}=1;
  }

  @scanners_to_check=grep($seen{$_},@scanners_to_check);
  return @scanners_to_check;
}


#################################################
# END of subroutines added by ST
#################################################

